# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import ndp_proxy as _ndp_proxy
from openstack.network.v2 import network
from openstack.network.v2 import port
from openstack.network.v2 import router
from openstack.network.v2 import subnet
from openstack.tests.functional import base


class TestNDPProxy(base.BaseFunctionalTest):
    IPV6 = 6
    EXT_CIDR = "2002::1:0/112"
    INT_CIDR = "2002::2:0/112"
    EXT_NET_ID: str
    INT_NET_ID: str
    EXT_SUB_ID: str
    INT_SUB_ID: str
    ROT_ID: str
    INTERNAL_PORT_ID: str

    def setUp(self):
        super().setUp()

        if not self.user_cloud.network.find_extension("l3-ndp-proxy"):
            self.skipTest("L3 ndp proxy extension disabled")

        self.ROT_NAME = self.getUniqueString()
        self.EXT_NET_NAME = self.getUniqueString()
        self.EXT_SUB_NAME = self.getUniqueString()
        self.INT_NET_NAME = self.getUniqueString()
        self.INT_SUB_NAME = self.getUniqueString()

        # Find External Network
        for net in self.user_cloud.network.networks(is_router_external=True):
            self.EXT_NET_ID = net.id
        # Find subnet of the chosen external net
        for sub in self.user_cloud.network.subnets(network_id=self.EXT_NET_ID):
            self.EXT_SUB_ID = sub.id
        if not self.EXT_NET_ID and self.operator_cloud:
            # There is no existing external net, but operator
            # credentials available
            # WARNING: this external net is not dropped
            # Create External Network
            net = self._create_network(
                self.EXT_NET_NAME, **{"router:external": True}
            )
            self.EXT_NET_ID = net.id
            sub = self._create_subnet(
                self.EXT_SUB_NAME, self.EXT_NET_ID, self.EXT_CIDR
            )
            self.EXT_SUB_ID = sub.id

        # Create Router
        sot = self.user_cloud.network.create_router(
            name=self.ROT_NAME,
            **{
                "external_gateway_info": {"network_id": self.EXT_NET_ID},
                "enable_ndp_proxy": True,
            },
        )
        assert isinstance(sot, router.Router)
        self.assertEqual(self.ROT_NAME, sot.name)
        self.ROT_ID = sot.id
        self.ROT = sot

        # Add Router's Interface to Internal Network
        interface = self.ROT.add_interface(
            self.user_cloud.network, subnet_id=self.INT_SUB_ID
        )
        self.assertEqual(interface["subnet_id"], self.INT_SUB_ID)

        # Create Port in Internal Network
        prt = self.user_cloud.network.create_port(network_id=self.INT_NET_ID)
        assert isinstance(prt, port.Port)
        self.INTERNAL_PORT_ID = prt.id
        self.INTERNAL_IP_ADDRESS = prt.fixed_ips[0]["ip_address"]

        # Create ndp proxy
        np = self.user_cloud.network.create_ndp_proxy(
            router_id=self.ROT_ID, port_id=self.INTERNAL_PORT_ID
        )
        assert isinstance(np, _ndp_proxy.NDPProxy)
        self.NP = np

    def tearDown(self):
        sot = self.user_cloud.network.delete_ndp_proxy(
            self.NP.id, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_port(
            self.INTERNAL_PORT_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.ROT.remove_interface(
            self.user_cloud.network, subnet_id=self.INT_SUB_ID
        )
        self.assertEqual(sot["subnet_id"], self.INT_SUB_ID)
        sot = self.user_cloud.network.delete_router(
            self.ROT_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_subnet(
            self.INT_SUB_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_network(
            self.INT_NET_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        super().tearDown()

    def _create_network(self, name, **args):
        self.name = name
        net = self.user_cloud.network.create_network(name=name, **args)
        assert isinstance(net, network.Network)
        self.assertEqual(self.name, net.name)
        return net

    def _create_subnet(self, name, net_id, cidr):
        self.name = name
        self.net_id = net_id
        self.cidr = cidr
        sub = self.user_cloud.network.create_subnet(
            name=self.name,
            ip_version=self.IPV6,
            network_id=self.net_id,
            cidr=self.cidr,
        )
        assert isinstance(sub, subnet.Subnet)
        self.assertEqual(self.name, sub.name)
        return sub

    def test_find(self):
        sot = self.user_cloud.network.find_ndp_proxy(
            self.NP.id, ignore_missing=False
        )
        self.assertEqual(self.ROT_ID, sot.router_id)
        self.assertEqual(self.INTERNAL_PORT_ID, sot.port_id)
        self.assertEqual(self.INTERNAL_IP_ADDRESS, sot.ip_address)

    def test_get(self):
        sot = self.user_cloud.network.get_ndp_proxy(self.NP.id)
        self.assertEqual(self.ROT_ID, sot.router_id)
        self.assertEqual(self.INTERNAL_PORT_ID, sot.port_id)
        self.assertEqual(self.INTERNAL_IP_ADDRESS, sot.ip_address)

    def test_list(self):
        np_ids = [o.id for o in self.user_cloud.network.ndp_proxies()]
        self.assertIn(self.NP.id, np_ids)

    def test_update(self):
        description = "balabalbala"
        sot = self.user_cloud.network.update_ndp_proxy(
            self.NP.id, description=description
        )
        self.assertEqual(description, sot.description)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import registered_limit as _registered_limit
from openstack.tests.functional.identity.v3 import base


class TestRegisteredLimit(base.BaseIdentityTest):
    def setUp(self):
        super().setUp()

        self.region_name = self.getUniqueString('region')
        self.region = self.system_admin_identity_client.create_region(
            name=self.region_name
        )
        self.addCleanup(
            self.system_admin_identity_client.delete_region, self.region
        )

        self.service_name = self.getUniqueString('service')
        self.service_type = self.getUniqueString('type')
        self.service = self.system_admin_identity_client.create_service(
            name=self.service_name,
            type=self.service_type,
        )
        self.addCleanup(
            self.system_admin_identity_client.delete_service, self.service
        )

        self.resource_name = self.getUniqueString('resource')
        self.registered_limit_description = self.getUniqueString(
            'registered_limit'
        )

    def _delete_registered_limit(self, registered_limit):
        ret = self.system_admin_identity_client.delete_registered_limit(
            registered_limit
        )
        self.assertIsNone(ret)

    def test_registered_limit(self):
        # create the registered limit

        registered_limit = (
            self.system_admin_identity_client.create_registered_limit(
                resource_name=self.resource_name,
                service_id=self.service.id,
                region_id=self.region.id,
                default_limit=10,
            )
        )
        self.addCleanup(self._delete_registered_limit, registered_limit)
        self.assertIsInstance(
            registered_limit, _registered_limit.RegisteredLimit
        )
        self.assertIsNotNone(registered_limit.id)
        self.assertIsNone(registered_limit.description)
        self.assertEqual(self.service.id, registered_limit.service_id)
        self.assertEqual(self.region.id, registered_limit.region_id)

        # update the registered limit

        registered_limit = (
            self.system_admin_identity_client.update_registered_limit(
                registered_limit, description=self.registered_limit_description
            )
        )
        self.assertIsInstance(
            registered_limit, _registered_limit.RegisteredLimit
        )
        self.assertEqual(
            self.registered_limit_description, registered_limit.description
        )

        # retrieve details of the (updated) registered limit by ID

        registered_limit = (
            self.system_admin_identity_client.get_registered_limit(
                registered_limit.id
            )
        )
        self.assertIsInstance(
            registered_limit, _registered_limit.RegisteredLimit
        )
        self.assertEqual(
            self.registered_limit_description, registered_limit.description
        )

        # (there's no name, so no way to retrieve by name)

        # list all registered limits

        registered_limits = list(
            self.system_admin_identity_client.registered_limits()
        )
        self.assertIsInstance(
            registered_limits[0], _registered_limit.RegisteredLimit
        )
        self.assertIn(
            self.resource_name,
            {x.resource_name for x in registered_limits},
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Generator
from typing import Any, Self

from keystoneauth1 import adapter

from openstack import resource


class Version(resource.Resource):
    resource_key = 'version'
    resources_key = 'versions'
    base_path = '/'

    # capabilities
    allow_list = True

    # Properties
    media_types = resource.Body('media-types')
    status = resource.Body('status')
    updated = resource.Body('updated')

    @classmethod
    def list(
        cls,
        session: adapter.Adapter,
        paginated: bool = True,
        base_path: str | None = None,
        allow_unknown_params: bool = False,
        *,
        microversion: str | None = None,
        headers: dict[str, str] | None = None,
        **params: Any,
    ) -> Generator[Self, None, None]:
        if base_path is None:
            base_path = cls.base_path

        resp = session.get(base_path, params=params)
        resp = resp.json()
        for data in resp[cls.resources_key]['values']:
            yield cls.existing(**data)



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack._hacking import checks
from openstack.tests.unit import base


class HackingTestCase(base.TestCase):
    """This class tests the hacking checks in openstack._hacking.checks.

    It works by passing strings to the check methods like the pep8/flake8
    parser would. The parser loops over each line in the file and then passes
    the parameters to the check method. The parameter names in the check method
    dictate what type of object is passed to the check method.

    The parameter types are::

        logical_line: A processed line with the following modifications:
            - Multi-line statements converted to a single line.
            - Stripped left and right.
            - Contents of strings replaced with "xxx" of same length.
            - Comments removed.
        physical_line: Raw line of text from the input file.
        lines: a list of the raw lines from the input file
        tokens: the tokens that contribute to this logical line
        line_number: line number in the input file
        total_lines: number of lines in the input file
        blank_lines: blank lines before this one
        indent_char: indentation character in this file (" " or "\t")
        indent_level: indentation (with tabs expanded to multiples of 8)
        previous_indent_level: indentation on previous line
        previous_logical: previous logical line
        filename: Path of the file being run through pep8

    When running a test on a check method the return will be False/None if
    there is no violation in the sample input. If there is an error a tuple is
    returned with a position in the line, and a message. So to check the result
    just assertTrue if the check is expected to fail and assertFalse if it
    should pass.
    """

    def test_assert_no_setupclass(self):
        self.assertEqual(
            len(list(checks.assert_no_setupclass("def setUpClass(cls)"))), 1
        )

        self.assertEqual(
            len(list(checks.assert_no_setupclass("# setUpClass is evil"))), 0
        )

        self.assertEqual(
            len(
                list(
                    checks.assert_no_setupclass(
                        "def setUpClassyDrinkingLocation(cls)"
                    )
                )
            ),
            0,
        )

    def test_assert_no_deprecated_exceptions(self):
        self.assertEqual(
            len(
                list(
                    checks.assert_no_deprecated_exceptions(
                        "raise exc.OpenStackCloudTimeout",
                        "openstack/cloud/compute.py",
                    )
                )
            ),
            1,
        )

        self.assertEqual(
            len(
                list(
                    checks.assert_no_deprecated_exceptions(
                        "raise exc.OpenStackCloudTimeout",
                        "openstack/cloud/exc.py",
                    )
                )
            ),
            0,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack import exceptions
from openstack.image.v2 import cache
from openstack.tests.unit import base


EXAMPLE = {
    'cached_images': [
        {
            'hits': 0,
            'image_id': '1a56983c-f71f-490b-a7ac-6b321a18935a',
            'last_accessed': 1671699579.444378,
            'last_modified': 1671699579.444378,
            'size': 0,
        },
    ],
    'queued_images': [
        '3a4560a1-e585-443e-9b39-553b46ec92d1',
        '6f99bf80-2ee6-47cf-acfe-1f1fabb7e810',
    ],
}


class TestCache(base.TestCase):
    def test_basic(self):
        sot = cache.Cache()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('/cache', sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_delete)

    def test_make_it(self):
        sot = cache.Cache(**EXAMPLE)
        self.assertEqual(
            [cache.CachedImage(**e) for e in EXAMPLE['cached_images']],
            sot.cached_images,
        )
        self.assertEqual(EXAMPLE['queued_images'], sot.queued_images)

    @mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
    def test_queue(self):
        sot = cache.Cache()
        sess = mock.Mock()
        sess.put = mock.Mock()
        sess.default_microversion = '2.14'

        sot.queue(sess, image='image_id')

        sess.put.assert_called_with(
            'cache/image_id', microversion=sess.default_microversion
        )

    @mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
    def test_clear(self):
        sot = cache.Cache(**EXAMPLE)
        session = mock.Mock()
        session.delete = mock.Mock()

        sot.clear(session)
        session.delete.assert_called_with('/cache', headers={})

        sot.clear(session, 'both')
        session.delete.assert_called_with('/cache', headers={})

        sot.clear(session, 'cache')
        session.delete.assert_called_with(
            '/cache', headers={'x-image-cache-clear-target': 'cache'}
        )

        sot.clear(session, 'queue')
        session.delete.assert_called_with(
            '/cache', headers={'x-image-cache-clear-target': 'queue'}
        )

        self.assertRaises(
            exceptions.InvalidRequest, sot.clear, session, 'invalid'
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Iterable
from typing import Any

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils


class QoSSpec(resource.Resource):
    resource_key = "qos_specs"
    resources_key = "qos_specs"
    base_path = "/qos-specs"

    _query_mapping = resource.QueryParameters(
        "project_id",
        "limit",
        "marker",
        "offset",
        "sort_dir",
        "sort_key",
        "sort",
    )

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # The Cinder API sends QoS spec properties as flat top-level keys in the
    # request body alongside 'name' and 'consumer', so we need to allow
    # unknown body attributes to be passed through.
    _allow_unknown_attrs_in_body = True

    # Properties
    consumer = resource.Body("consumer")
    id = resource.Body("id", type=str)
    name = resource.Body("name")
    specs = resource.Body("specs")

    def associate(self, session: adapter.Adapter, vol_type_id: str) -> None:
        """Associate this QoS spec with a volume type.

        :param session: The session to use for making this request.
        :param vol_type_id: The ID of the volume type to associate with.
        """
        url = utils.urljoin(self.base_path, self.id, 'associate')
        resp = session.get(url, params={'vol_type_id': vol_type_id})
        exceptions.raise_from_response(resp)

    def disassociate(self, session: adapter.Adapter, vol_type_id: str) -> None:
        """Disassociate this QoS spec from a volume type.

        :param session: The session to use for making this request.
        :param vol_type_id: The ID of the volume type to disassociate from.
        """
        url = utils.urljoin(self.base_path, self.id, 'disassociate')
        resp = session.get(url, params={'vol_type_id': vol_type_id})
        exceptions.raise_from_response(resp)

    def disassociate_all(self, session: adapter.Adapter) -> None:
        """Disassociate this QoS spec from all volume types.

        :param session: The session to use for making this request.
        """
        url = utils.urljoin(self.base_path, self.id, 'disassociate_all')
        resp = session.get(url)
        exceptions.raise_from_response(resp)

    def delete_keys(
        self, session: adapter.Adapter, keys: Iterable[Any]
    ) -> None:
        """Delete keys from this QoS spec.

        :param session: The session to use for making this request.
        :param keys: The keys to delete.
        """
        url = utils.urljoin(self.base_path, self.id, 'delete_keys')
        resp = session.put(url, json={'keys': keys})
        exceptions.raise_from_response(resp)


class QoSSpecAssociation(resource.Resource):
    resources_key = "qos_associations"
    base_path = "/qos-specs/%(qos_spec_id)s/associations"

    # capabilities
    allow_list = True

    # Properties
    #: The type of association (e.g. "volume_type").
    association_type = resource.Body("association_type")
    #: The ID of the associated resource.
    id = resource.Body("id", type=str)
    #: The name of the associated resource.
    name = resource.Body("name")
    #: The ID of the QoS spec.
    qos_spec_id = resource.URI("qos_spec_id")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any

from keystoneauth1 import adapter

from openstack import resource
from openstack import utils


class Flavor(resource.Resource):
    resource_key = 'flavor'
    resources_key = 'flavors'
    base_path = '/flavors'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'description',
        'name',
        'service_type',
        'sort_key',
        'sort_dir',
        is_enabled='enabled',
    )

    # properties
    #: description for the flavor
    description = resource.Body('description')
    #: Sets enabled flag
    is_enabled = resource.Body('enabled', type=bool)
    #: The name of the flavor
    name = resource.Body('name')
    #: Service type to which the flavor applies
    service_type = resource.Body('service_type')
    #: IDs of service profiles associated with this flavor
    service_profile_ids = resource.Body('service_profiles', type=list)

    def associate_flavor_with_service_profile(
        self,
        session: adapter.Adapter,
        service_profile_id: str | None = None,
    ) -> dict[str, Any]:
        flavor_id = self.id
        url = utils.urljoin(self.base_path, flavor_id, 'service_profiles')
        body = {"service_profile": {"id": service_profile_id}}
        resp = session.post(url, json=body)
        return resp.json()  # type: ignore[no-any-return]

    def disassociate_flavor_from_service_profile(
        self,
        session: adapter.Adapter,
        service_profile_id: str | None = None,
    ) -> None:
        flavor_id = self.id
        url = utils.urljoin(
            self.base_path, flavor_id, 'service_profiles', service_profile_id
        )
        session.delete(
            url,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.baremetal.v1 import inspection_rules
from openstack.tests.unit import base


FAKE = {
    "created_at": "2025-03-18T22:28:48.643434+11:11",
    "description": "BMC credentials",
    "phase": "main",
    "priority": 100,
    "sensitive": False,
    "actions": [
        {
            "op": "set-attribute",
            "args": {
                "path": "/properties/cpus",
                "value": "{inventory[cpu][count]}",
            },
        },
        {
            "op": "set-attribute",
            "args": {
                "path": "/properties/memory_mb",
                "value": "{inventory[memory][physical_mb]}",
            },
        },
        {
            "op": "set-attribute",
            "args": {
                "path": "/properties/cpu_arch",
                "value": "{inventory[cpu][architecture]}",
            },
        },
    ],
    "conditions": [
        {"op": "is-true", "args": {"value": "{inventory[cpu][count]}"}}
    ],
    "links": [
        {
            "href": "http://10.60.253.180:6385/v1/inspection_rules"
            "/783bf33a-a8e3-1e23-a645-1e95a1f95186",
            "rel": "self",
        },
        {
            "href": "http://10.60.253.180:6385/inspection_rules"
            "/783bf33a-a8e3-1e23-a645-1e95a1f95186",
            "rel": "bookmark",
        },
    ],
    "updated_at": None,
    "uuid": "783bf33a-a8e3-1e23-a645-1e95a1f95186",
}


class InspectionRules(base.TestCase):
    def test_basic(self):
        sot = inspection_rules.InspectionRule()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('inspection_rules', sot.resources_key)
        self.assertEqual('/inspection_rules', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual('PATCH', sot.commit_method)

    def test_instantiate(self):
        sot = inspection_rules.InspectionRule(**FAKE)
        self.assertEqual(FAKE['actions'], sot.actions)
        self.assertEqual(FAKE['description'], sot.description)
        self.assertEqual(FAKE['conditions'], sot.conditions)
        self.assertEqual(FAKE['created_at'], sot.created_at)
        self.assertEqual(FAKE['links'], sot.links)
        self.assertEqual(FAKE['phase'], sot.phase)
        self.assertEqual(FAKE['priority'], sot.priority)
        self.assertEqual(FAKE['sensitive'], sot.sensitive)
        self.assertEqual(FAKE['updated_at'], sot.updated_at)
        self.assertEqual(FAKE['uuid'], sot.id)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack.cloud import _utils
from openstack.cloud import openstackcloud
from openstack import exceptions
from openstack.orchestration.util import event_utils


class OrchestrationCloudMixin(openstackcloud._OpenStackCloudMixin):
    def get_template_contents(
        self,
        template_file=None,
        template_url=None,
        template_object=None,
        files=None,
    ):
        return self.orchestration.get_template_contents(
            template_file=template_file,
            template_url=template_url,
            template_object=template_object,
            files=files,
        )

    def create_stack(
        self,
        name,
        tags=None,
        template_file=None,
        template_url=None,
        template_object=None,
        files=None,
        rollback=True,
        wait=False,
        timeout=3600,
        environment_files=None,
        **parameters,
    ):
        """Create a stack.

        :param string name: Name of the stack.
        :param tags: List of tag(s) of the stack. (optional)
        :param string template_file: Path to the template.
        :param string template_url: URL of template.
        :param string template_object: URL to retrieve template object.
        :param dict files: dict of additional file content to include.
        :param boolean rollback: Enable rollback on create failure.
        :param boolean wait: Whether to wait for the delete to finish.
        :param int timeout: Stack create timeout in seconds.
        :param environment_files: Paths to environment files to apply.

        Other arguments will be passed as stack parameters which will take
        precedence over any parameters specified in the environments.

        Only one of template_file, template_url, template_object should be
        specified.

        :returns: a dict containing the stack description
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call
        """
        params = dict(
            tags=tags,
            is_rollback_disabled=not rollback,
            timeout_mins=timeout // 60,
            parameters=parameters,
        )
        params.update(
            self.orchestration.read_env_and_templates(
                template_file=template_file,
                template_url=template_url,
                template_object=template_object,
                files=files,
                environment_files=environment_files,
            )
        )
        self.orchestration.create_stack(name=name, **params)
        if wait:
            event_utils.poll_for_events(self, stack_name=name, action='CREATE')
        return self.get_stack(name)

    def update_stack(
        self,
        name_or_id,
        template_file=None,
        template_url=None,
        template_object=None,
        files=None,
        rollback=True,
        tags=None,
        wait=False,
        timeout=3600,
        environment_files=None,
        **parameters,
    ):
        """Update a stack.

        :param string name_or_id: Name or ID of the stack to update.
        :param string template_file: Path to the template.
        :param string template_url: URL of template.
        :param string template_object: URL to retrieve template object.
        :param dict files: dict of additional file content to include.
        :param boolean rollback: Enable rollback on update failure.
        :param boolean wait: Whether to wait for the delete to finish.
        :param int timeout: Stack update timeout in seconds.
        :param environment_files: Paths to environment files to apply.

        Other arguments will be passed as stack parameters which will take
        precedence over any parameters specified in the environments.

        Only one of template_file, template_url, template_object should be
        specified.

        :returns: a dict containing the stack description
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API calls
        """
        params = dict(
            tags=tags,
            is_rollback_disabled=not rollback,
            timeout_mins=timeout // 60,
            parameters=parameters,
        )
        params.update(
            self.orchestration.read_env_and_templates(
                template_file=template_file,
                template_url=template_url,
                template_object=template_object,
                files=files,
                environment_files=environment_files,
            )
        )
        if wait:
            # find the last event to use as the marker
            events = event_utils.get_events(
                self, name_or_id, event_args={'sort_dir': 'desc', 'limit': 1}
            )
            marker = events[0].id if events else None

        self.orchestration.update_stack(name_or_id, **params)

        if wait:
            event_utils.poll_for_events(
                self, name_or_id, action='UPDATE', marker=marker
            )
        return self.get_stack(name_or_id)

    def delete_stack(self, name_or_id, wait=False):
        """Delete a stack

        :param string name_or_id: Stack name or ID.
        :param boolean wait: Whether to wait for the delete to finish

        :returns: True if delete succeeded, False if the stack was not found.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call
        """
        stack = self.get_stack(name_or_id, resolve_outputs=False)
        if stack is None:
            self.log.debug("Stack %s not found for deleting", name_or_id)
            return False

        if wait:
            # find the last event to use as the marker
            events = event_utils.get_events(
                self, name_or_id, event_args={'sort_dir': 'desc', 'limit': 1}
            )
            marker = events[0].id if events else None

        self.orchestration.delete_stack(stack)

        if wait:
            try:
                event_utils.poll_for_events(
                    self, stack_name=name_or_id, action='DELETE', marker=marker
                )
            except exceptions.HttpException:
                pass
            stack = self.get_stack(name_or_id, resolve_outputs=False)
            if stack and stack['stack_status'] == 'DELETE_FAILED':
                raise exceptions.SDKException(
                    "Failed to delete stack {id}: {reason}".format(
                        id=name_or_id, reason=stack['stack_status_reason']
                    )
                )

        return True

    def search_stacks(self, name_or_id=None, filters=None):
        """Search stacks.

        :param name_or_id: Name or ID of the desired stack.
        :param filters: a dict containing additional filters to use. e.g.
                {'stack_status': 'CREATE_COMPLETE'}

        :returns: a list of ``openstack.orchestration.v1.stack.Stack``
            containing the stack description.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        stacks = self.list_stacks()
        return _utils._filter_list(stacks, name_or_id, filters)

    def list_stacks(self, **query):
        """List all stacks.

        :param dict query: Query parameters to limit stacks.

        :returns: a list of :class:`openstack.orchestration.v1.stack.Stack`
            objects containing the stack description.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        return list(self.orchestration.stacks(**query))

    def get_stack(self, name_or_id, filters=None, resolve_outputs=True):
        """Get exactly one stack.

        :param name_or_id: Name or ID of the desired stack.
        :param filters: a dict containing additional filters to use. e.g.
                {'stack_status': 'CREATE_COMPLETE'}
        :param resolve_outputs: If True, then outputs for this
                stack will be resolved

        :returns: a :class:`openstack.orchestration.v1.stack.Stack`
            containing the stack description
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call or if multiple matches are
            found.
        """

        def _search_one_stack(name_or_id=None, filters=None):
            # stack names are mandatory and enforced unique in the project
            # so a StackGet can always be used for name or ID.
            try:
                stack = self.orchestration.find_stack(
                    name_or_id,
                    ignore_missing=False,
                    resolve_outputs=resolve_outputs,
                )
                if stack.status == 'DELETE_COMPLETE':
                    return []
            except exceptions.NotFoundException:
                return []
            return _utils._filter_list([stack], name_or_id, filters)

        return _utils._get_entity(self, _search_one_stack, name_or_id, filters)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.block_storage.v3 import cluster
from openstack.tests.unit import base

EXAMPLE = {
    "binary": "cinder-volume",
    "name": "cluster_name",
    "state": "up",
    "status": "enabled",
    "replication_status": None,
    "num_hosts": 3,
    "num_down_hosts": 0,
    "last_heartbeat": "2017-06-29T05:50:35.000000",
    "created_at": "2017-06-29T05:50:35.000000",
    "updated_at": "2017-06-29T05:50:35.000000",
    "disabled_reason": None,
    "frozen": False,
    "active_backend_id": None,
}


class TestCluster(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.json = mock.Mock(return_value={'cluster': {}})
        self.resp.status_code = 200
        self.resp.headers = {}
        self.sess = mock.Mock()
        self.sess.put = mock.Mock(return_value=self.resp)
        self.sess.default_microversion = '3.7'

    def test_basic(self):
        sot = cluster.Cluster()
        self.assertEqual('cluster', sot.resource_key)
        self.assertEqual('clusters', sot.resources_key)
        self.assertEqual('/clusters', sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_list)
        self.assertFalse(sot.allow_create)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)

        self.assertDictEqual(
            {
                'active_backend_id': 'active_backend_id',
                'binary': 'binary',
                'disabled': 'disabled',
                'frozen': 'frozen',
                'is_up': 'is_up',
                'limit': 'limit',
                'marker': 'marker',
                'name': 'name',
                'num_down_hosts': 'num_down_hosts',
                'num_hosts': 'num_hosts',
                'replication_status': 'replication_status',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = cluster.Cluster(**EXAMPLE)
        self.assertEqual(EXAMPLE['binary'], sot.binary)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['state'], sot.state)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['replication_status'], sot.replication_status)
        self.assertEqual(EXAMPLE['num_hosts'], sot.num_hosts)
        self.assertEqual(EXAMPLE['num_down_hosts'], sot.num_down_hosts)
        self.assertEqual(EXAMPLE['last_heartbeat'], sot.last_heartbeat)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
        self.assertEqual(EXAMPLE['disabled_reason'], sot.disabled_reason)
        self.assertEqual(EXAMPLE['frozen'], sot.frozen)
        self.assertEqual(EXAMPLE['active_backend_id'], sot.active_backend_id)

    def test_enable(self):
        sot = cluster.Cluster(**EXAMPLE)

        res = sot.enable(self.sess)
        self.assertIsNotNone(res)

        url = 'clusters/enable'
        body = {
            'name': 'cluster_name',
            'binary': 'cinder-volume',
        }
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    def test_disable(self):
        sot = cluster.Cluster(**EXAMPLE)

        res = sot.disable(self.sess)
        self.assertIsNotNone(res)

        url = 'clusters/disable'
        body = {
            'name': 'cluster_name',
            'binary': 'cinder-volume',
        }
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )

    def test_disable__with_reason(self):
        sot = cluster.Cluster(**EXAMPLE)
        reason = 'for testing'

        res = sot.disable(self.sess, reason=reason)
        self.assertIsNotNone(res)

        url = 'clusters/disable'
        body = {
            'name': 'cluster_name',
            'binary': 'cinder-volume',
            'disabled_reason': reason,
        }
        self.sess.put.assert_called_with(
            url,
            json=body,
            microversion=self.sess.default_microversion,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import adapter

from openstack.common import tag
from openstack import resource
from openstack import utils


class Domain(resource.Resource, tag.TagMixin):
    resource_key = 'domain'
    resources_key = 'domains'
    base_path = '/domains'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    commit_method = 'PATCH'

    _query_mapping = resource.QueryParameters(
        'name',
        is_enabled='enabled',
        **tag.TagMixin._tag_query_parameters,
    )

    # Properties
    #: The description of this domain. *Type: string*
    description = resource.Body('description')
    #: Setting this attribute to ``False`` prevents users from authorizing
    #: against this domain or any projects owned by this domain, and prevents
    #: users owned by this domain from authenticating or receiving any other
    #: authorization. Additionally, all pre-existing tokens applicable
    #: to the above entities are immediately invalidated.
    #: Re-enabling a domain does not re-enable pre-existing tokens.
    #: *Type: bool*
    is_enabled = resource.Body('enabled', type=bool)
    #: The resource options for the project. Available resource options are
    #: immutable.
    options = resource.Body('options', type=dict)
    #: The links related to the domain resource.
    links = resource.Body('links')

    def assign_role_to_user(
        self,
        session: adapter.Adapter,
        user: resource.Resource,
        role: resource.Resource,
        inherited: bool,
    ) -> bool:
        """Assign role to user on domain"""
        url = utils.urljoin(
            self.base_path,
            self.id,
            'users',
            user.id,
            'roles',
            role.id,
        )
        if inherited:
            url = utils.urljoin('OS-INHERIT', url, 'inherited_to_projects')
        resp = session.put(
            url,
        )
        if resp.status_code == 204:
            return True
        return False

    def validate_user_has_role(
        self,
        session: adapter.Adapter,
        user: resource.Resource,
        role: resource.Resource,
        inherited: bool,
    ) -> bool:
        """Validates that a user has a role on a domain"""
        url = utils.urljoin(
            self.base_path, self.id, 'users', user.id, 'roles', role.id
        )
        if inherited:
            url = utils.urljoin('OS-INHERIT', url, 'inherited_to_projects')
        resp = session.head(
            url,
        )
        if resp.status_code == 204:
            return True
        return False

    def unassign_role_from_user(
        self,
        session: adapter.Adapter,
        user: resource.Resource,
        role: resource.Resource,
        inherited: bool,
    ) -> bool:
        """Unassigns a role from a user on a domain"""
        url = utils.urljoin(
            self.base_path, self.id, 'users', user.id, 'roles', role.id
        )
        if inherited:
            url = utils.urljoin('OS-INHERIT', url, 'inherited_to_projects')
        resp = session.delete(
            url,
        )
        if resp.status_code == 204:
            return True
        return False

    def assign_role_to_group(
        self,
        session: adapter.Adapter,
        group: resource.Resource,
        role: resource.Resource,
        inherited: bool,
    ) -> bool:
        """Assign role to group on domain"""
        url = utils.urljoin(
            self.base_path,
            self.id,
            'groups',
            group.id,
            'roles',
            role.id,
        )
        if inherited:
            url = utils.urljoin('OS-INHERIT', url, 'inherited_to_projects')
        resp = session.put(
            url,
        )
        if resp.status_code == 204:
            return True
        return False

    def validate_group_has_role(
        self,
        session: adapter.Adapter,
        group: resource.Resource,
        role: resource.Resource,
        inherited: bool,
    ) -> bool:
        """Validates that a group has a role on a domain"""
        url = utils.urljoin(
            self.base_path, self.id, 'groups', group.id, 'roles', role.id
        )
        if inherited:
            url = utils.urljoin('OS-INHERIT', url, 'inherited_to_projects')
        resp = session.head(
            url,
        )
        if resp.status_code == 204:
            return True
        return False

    def unassign_role_from_group(
        self,
        session: adapter.Adapter,
        group: resource.Resource,
        role: resource.Resource,
        inherited: bool,
    ) -> bool:
        """Unassigns a role from a group on a domain"""
        url = utils.urljoin(
            self.base_path, self.id, 'groups', group.id, 'roles', role.id
        )
        if inherited:
            url = utils.urljoin('OS-INHERIT', url, 'inherited_to_projects')
        resp = session.delete(
            url,
        )
        if resp.status_code == 204:
            return True
        return False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Generator, Iterable
from typing import Any, Literal, Self, overload

from keystoneauth1 import adapter

from openstack import fields
from openstack import resource


RETRIABLE_STATUS_CODES = [
    # HTTP Conflict - happens if a node is locked
    409,
    # HTTP Service Unavailable happens if there's no free conductor
    503,
]
"""HTTP status codes that should be retried."""


PROVISIONING_VERSIONS = {
    'abort': 13,
    'adopt': 17,
    'clean': 15,
    'inspect': 6,
    'manage': 4,
    'provide': 4,
    'rescue': 38,
    'unrescue': 38,
    'unhold': 85,
    'service': 87,
}
"""API microversions introducing provisioning verbs."""


# Based on https://docs.openstack.org/ironic/latest/contributor/states.html
EXPECTED_STATES = {
    'active': 'active',
    'adopt': 'available',
    'clean': 'manageable',
    'deleted': 'available',
    'inspect': 'manageable',
    'manage': 'manageable',
    'provide': 'available',
    'rebuild': 'active',
    'rescue': 'rescue',
}
"""Mapping of provisioning actions to expected stable states."""

EXPECTED_POWER_STATES = {
    'power on': 'power on',
    'power off': 'power off',
    'rebooting': 'power on',
    'soft power off': 'power off',
    'soft rebooting': 'power on',
}
"""Mapping of target power states to expected power states."""

STATE_VERSIONS = {
    'available': '1.1',
    'enroll': '1.11',
    'manageable': '1.4',
}
"""API versions when certain states were introduced."""

VIF_VERSION = '1.28'
"""API version in which the VIF operations were introduced."""

VIF_OPTIONAL_PARAMS_VERSION = '1.67'
"""API version in which the VIF optional parameters were introduced."""

INJECT_NMI_VERSION = '1.29'
"""API vresion in which support for injecting NMI was introduced."""

CONFIG_DRIVE_REBUILD_VERSION = '1.35'
"""API version in which rebuild accepts a configdrive."""

RESET_INTERFACES_VERSION = '1.45'
"""API version in which the reset_interfaces parameter was introduced."""

CONFIG_DRIVE_DICT_VERSION = '1.56'
"""API version in which configdrive can be a dictionary."""

DEPLOY_STEPS_VERSION = '1.69'
"""API version in which deploy_steps was added to node provisioning."""

CHANGE_BOOT_MODE_VERSION = '1.76'
"""API version in which boot_mode and secure_boot states can be changed"""

FIRMWARE_VERSION = '1.86'
"""API version in which firmware components of a node can be accessed"""

VMEDIA_VERSION = '1.89'
"""API version in which the virtual media operations were introduced."""

RUNBOOKS_VERSION = '1.92'
"""API version in which a runbook can be used in place of arbitrary steps
for provisioning"""


class Resource(resource.Resource):
    """A subclass for resources that use the path to request a detailed view.

    Two patterns exist for fetching the detailed view when listing resources.

    - As part of the path. For example:

        GET /v1/ports/detail

    - As a query parameter. For example:

        GET /v1/conductors?detail=True

    This handles resources that use the former pattern, namely:

    - chassis
    - nodes
    - ports
    - portgroups
    """

    base_path: str

    @classmethod
    def list(
        cls,
        session: adapter.Adapter,
        paginated: bool = True,
        base_path: str | None = None,
        allow_unknown_params: bool = False,
        *,
        microversion: str | None = None,
        details: bool = False,
        **params: Any,
    ) -> Generator[Self, None, None]:
        """This method is a generator which yields resource objects.

        This resource object list generator handles pagination and takes query
        params for response filtering.

        :param session: The session to use for making this request.
        :param paginated: ``True`` if a GET to this resource returns
            a paginated series of responses, or ``False`` if a GET returns only
            one page of data. **When paginated is False only one page of data
            will be returned regardless of the API's support of pagination.**
        :param base_path: Base part of the URI for listing resources, if
            different from :data:`~openstack.resource.Resource.base_path`.
        :param allow_unknown_params: ``True`` to accept, but discard
            unknown query parameters. This allows getting list of 'filters' and
            passing everything known to the server. ``False`` will result in
            validation exception when unknown query parameters are passed.
        :param microversion: API version to override the negotiated one.
        :param details: Whether to return detailed resource records.
        :param params: These keyword arguments are passed through the
            :meth:`~openstack.resource.QueryParamter._transpose` method
            to find if any of them match expected query parameters to be sent
            in the *params* argument to
            :meth:`~keystoneauth1.adapter.Adapter.get`. They are additionally
            checked against the :data:`~openstack.resource.Resource.base_path`
            format string to see if any path fragments need to be filled in by
            the contents of this argument.
            Parameters supported as filters by the server side are passed in
            the API call, remaining parameters are applied as filters to the
            retrieved results.

        :return: A generator of :class:`Resource` objects.
        :raises: :exc:`~openstack.exceptions.MethodNotSupported` if
            :data:`Resource.allow_list` is not set to ``True``.
        :raises: :exc:`~openstack.exceptions.InvalidResourceQuery` if query
            contains invalid params.
        """
        if not base_path:
            base_path = cls.base_path
            if details:
                base_path += '/detail'

        return super().list(
            session,
            paginated=paginated,
            base_path=base_path,
            allow_unknown_params=allow_unknown_params,
            microversion=microversion,
            **params,
        )

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: Literal[False],
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        details: bool = False,
        **params: Any,
    ) -> Self: ...

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        details: bool = False,
        **params: Any,
    ) -> Self | None: ...

    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        details: bool = False,
        **params: Any,
    ) -> Self | None:
        """Find a resource by its name or id.

        :param session: The session to use for making this request.
        :param name_or_id: This resource's identifier, if needed by
            the request. The default is ``None``.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.ResourceNotFound` will be raised when
            the resource does not exist.  When set to ``True``, None will be
            returned when attempting to find a nonexistent resource.
        :param list_base_path: base_path to be used when need listing
            resources.
        :param microversion: API version to override the negotiated one.
        :param details: Whether to return detailed resource records.
        :param params: Any additional parameters to be passed into
            underlying methods, such as to
            :meth:`~openstack.resource.Resource.existing` in order to pass on
            URI parameters.

        :return: The :class:`Resource` object matching the given name or id
            or None if nothing matches.
        :raises: :class:`openstack.exceptions.DuplicateResource` if more
            than one resource is found for this request.
        :raises: :class:`openstack.exceptions.ResourceNotFound` if nothing
            is found and ignore_missing is ``False``.
        """
        if not list_base_path:
            list_base_path = cls.base_path
            if details:
                list_base_path += '/detail'

        return super().find(
            session,
            name_or_id,
            ignore_missing=ignore_missing,
            list_base_path=list_base_path,
            microversion=microversion,
            all_projects=all_projects,
            **params,
        )


def comma_separated_list(value: Iterable[str] | None) -> str | None:
    if value is None:
        return None
    else:
        return ','.join(value)


def fields_type(
    value: Iterable[str] | None,
    resource_type: type[resource.Resource],
) -> str | None:
    if value is None:
        return None

    resource_mapping = {
        key: value.name
        for key, value in resource_type.__dict__.items()
        if isinstance(value, fields.Body)
    }

    return comma_separated_list(resource_mapping.get(x, x) for x in value)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from unittest import mock

from keystoneauth1 import adapter

from openstack.identity.v3 import limit
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    "service_id": "8ac43bb0926245cead88676a96c750d3",
    "region_id": 'RegionOne',
    "resource_name": 'cores',
    "resource_limit": 10,
    "project_id": 'a8455cdd4249498f99b63d5af2fb4bc8',
    "description": "compute cores for project 123",
    "links": {"self": "http://example.com/v3/limit_1"},
}


class TestLimit(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.status_code = 200
        self.resp.headers = {}
        self.resp.json = mock.Mock(return_value={})
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.default_microversion = None
        self.sess.retriable_status_codes = None
        self.sess._get_connection = mock.Mock(return_value=self.cloud)

    def test_basic(self):
        sot = limit.Limit()
        self.assertEqual('limits', sot.resources_key)
        self.assertEqual('/limits', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual('PATCH', sot.commit_method)

        self.assertDictEqual(
            {
                'service_id': 'service_id',
                'region_id': 'region_id',
                'resource_name': 'resource_name',
                'project_id': 'project_id',
                'marker': 'marker',
                'limit': 'limit',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = limit.Limit(**EXAMPLE)
        self.assertEqual(EXAMPLE['service_id'], sot.service_id)
        self.assertEqual(EXAMPLE['region_id'], sot.region_id)
        self.assertEqual(EXAMPLE['resource_name'], sot.resource_name)
        self.assertEqual(EXAMPLE['resource_limit'], sot.resource_limit)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['links'], sot.links)

    def test_create(self):
        sot = limit.Limit(**EXAMPLE)
        sot._translate_response = mock.Mock()
        self.sess.post = mock.Mock(return_value=self.resp)

        sot.create(self.sess)

        self.sess.post.assert_called_with(
            '/limits',
            json={'limits': [copy.deepcopy(EXAMPLE)]},
            microversion=None,
            headers={},
            params={},
        )

    def test_commit(self):
        sot = limit.Limit(id=IDENTIFIER, resource_limit=20)
        self.sess.patch = mock.Mock(return_value=self.resp)

        sot.commit(self.sess)

        self.sess.patch.assert_called_with(
            f'limits/{IDENTIFIER}',
            json={'limit': {'resource_limit': 20}},
            headers={},
            microversion=None,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ConsistencyGroupSnapshot(resource.Resource):
    resource_key = "cgsnapshot"
    resources_key = "cgsnapshots"
    base_path = "/cgsnapshots"

    _query_mapping = resource.QueryParameters(
        "limit",
        "marker",
        "offset",
        "sort_dir",
        "sort_key",
        "sort",
    )

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_list = True

    # Properties
    #: The UUID of the source consistency group.
    consistencygroup_id = resource.Body("consistencygroup_id")
    #: The date and time when this resource was created.
    created_at = resource.Body("created_at")
    #: The description.
    description = resource.Body("description")
    #: The name.
    name = resource.Body("name")
    #: The status.
    status = resource.Body("status")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v3 import default_type as _default_type
from openstack.tests.functional.block_storage.v3 import base


class TestDefaultType(base.BaseBlockStorageTest):
    def setUp(self):
        super().setUp()
        if not self._operator_cloud_name:
            self.skipTest("Operator cloud must be set for this test")
        self._set_operator_cloud(block_storage_api_version='3.67')
        block_storage = self.operator_cloud.block_storage
        assert block_storage.api_version == '3'
        self.admin_block_storage_client = block_storage
        self.PROJECT_ID = self.create_temporary_project().id

    def test_default_type(self):
        # Create a volume type
        type_name = self.getUniqueString()
        volume_type_id = self.admin_block_storage_client.create_type(
            name=type_name,
        ).id

        # Set default type for a project
        default_type = self.admin_block_storage_client.set_default_type(
            self.PROJECT_ID,
            volume_type_id,
        )
        self.assertIsInstance(default_type, _default_type.DefaultType)

        # Show default type for a project
        default_type = self.admin_block_storage_client.show_default_type(
            self.PROJECT_ID
        )
        self.assertIsInstance(default_type, _default_type.DefaultType)
        self.assertEqual(volume_type_id, default_type.volume_type_id)

        # List all default types
        default_types = self.admin_block_storage_client.default_types()
        for default_type in default_types:
            self.assertIsInstance(default_type, _default_type.DefaultType)
            # There could be existing default types set in the environment
            # Just verify that the default type we have set is correct
            if self.PROJECT_ID == default_type.project_id:
                self.assertEqual(volume_type_id, default_type.volume_type_id)

        # Unset default type for a project
        self.admin_block_storage_client.unset_default_type(self.PROJECT_ID)

        # Delete the volume type
        vol_type = self.admin_block_storage_client.delete_type(
            volume_type_id,
            ignore_missing=False,
        )
        self.assertIsNone(vol_type)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack import exceptions

OpenStackCloudException = exceptions.SDKException


class OpenStackCloudUnavailableExtension(OpenStackCloudException):
    pass


class OpenStackCloudUnavailableFeature(OpenStackCloudException):
    pass


# Backwards compat. These are deprecated and should not be used in new code.
class OpenStackCloudCreateException(OpenStackCloudException):
    def __init__(self, resource, resource_id, extra_data=None, **kwargs):
        super().__init__(
            message=f"Error creating {resource}: {resource_id}",
            extra_data=extra_data,
            **kwargs,
        )
        self.resource_id = resource_id


OpenStackCloudTimeout = exceptions.ResourceTimeout
OpenStackCloudHTTPError = exceptions.HttpException
OpenStackCloudBadRequest = exceptions.BadRequestException
OpenStackCloudURINotFound = exceptions.NotFoundException
OpenStackCloudResourceNotFound = OpenStackCloudURINotFound



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.image.v2 import metadef_namespace as _metadef_namespace
from openstack.tests.functional.image.v2 import base


class TestMetadefNamespace(base.BaseImageTest):
    # TODO(stephenfin): We should use setUpClass here for MOAR SPEED!!!
    def setUp(self):
        super().setUp()

        # there's a limit on namespace length
        namespace = self.getUniqueString().split('.')[-1]
        self.metadef_namespace = (
            self.admin_image_client.create_metadef_namespace(
                namespace=namespace,
            )
        )
        self.assertIsInstance(
            self.metadef_namespace,
            _metadef_namespace.MetadefNamespace,
        )
        self.assertEqual(namespace, self.metadef_namespace.namespace)

    def tearDown(self):
        # we do this in tearDown rather than via 'addCleanup' since we want to
        # wait for the deletion of the resource to ensure it completes
        self.admin_image_client.delete_metadef_namespace(
            self.metadef_namespace
        )
        self.admin_image_client.wait_for_delete(self.metadef_namespace)

        super().tearDown()

    def test_metadef_namespace(self):
        # get
        metadef_namespace = self.admin_image_client.get_metadef_namespace(
            self.metadef_namespace.namespace
        )
        self.assertEqual(
            self.metadef_namespace.namespace,
            metadef_namespace.namespace,
        )

        # (no find_metadef_namespace method)

        # list
        metadef_namespaces = list(self.admin_image_client.metadef_namespaces())
        # there are a load of default metadef namespaces so we don't assert
        # that this is the *only* metadef namespace present
        self.assertIn(
            self.metadef_namespace.namespace,
            {n.namespace for n in metadef_namespaces},
        )

        # update
        # there's a limit on display name and description lengths and no
        # inherent need for randomness so we use fixed strings
        metadef_namespace_display_name = 'A display name'
        metadef_namespace_description = 'A description'
        metadef_namespace = self.admin_image_client.update_metadef_namespace(
            self.metadef_namespace,
            display_name=metadef_namespace_display_name,
            description=metadef_namespace_description,
        )
        self.assertIsInstance(
            metadef_namespace,
            _metadef_namespace.MetadefNamespace,
        )
        metadef_namespace = self.admin_image_client.get_metadef_namespace(
            self.metadef_namespace.namespace
        )
        self.assertEqual(
            metadef_namespace_display_name,
            metadef_namespace.display_name,
        )
        self.assertEqual(
            metadef_namespace_description,
            metadef_namespace.description,
        )

    def test_tags(self):
        # add tag
        metadef_namespace = self.admin_image_client.get_metadef_namespace(
            self.metadef_namespace.namespace
        )
        metadef_namespace.add_tag(self.admin_image_client, 't1')
        metadef_namespace.add_tag(self.admin_image_client, 't2')

        # list tags
        metadef_namespace.fetch_tags(self.admin_image_client)
        md_tags = [tag['name'] for tag in metadef_namespace.tags]
        self.assertIn('t1', md_tags)
        self.assertIn('t2', md_tags)

        # remove tag
        metadef_namespace.remove_tag(self.admin_image_client, 't1')
        metadef_namespace = self.admin_image_client.get_metadef_namespace(
            self.metadef_namespace.namespace
        )
        md_tags = [tag['name'] for tag in metadef_namespace.tags]
        self.assertIn('t2', md_tags)
        self.assertNotIn('t1', md_tags)

        # add tags without append
        metadef_namespace.set_tags(self.admin_image_client, ["t1", "t2"])
        metadef_namespace.fetch_tags(self.admin_image_client)
        md_tags = [tag['name'] for tag in metadef_namespace.tags]
        self.assertIn('t1', md_tags)
        self.assertIn('t2', md_tags)

        # add tags with append
        metadef_namespace.set_tags(
            self.admin_image_client, ["t3", "t4"], append=True
        )
        metadef_namespace.fetch_tags(self.admin_image_client)
        md_tags = [tag['name'] for tag in metadef_namespace.tags]
        self.assertIn('t1', md_tags)
        self.assertIn('t2', md_tags)
        self.assertIn('t3', md_tags)
        self.assertIn('t4', md_tags)

        # remove all tags
        metadef_namespace.remove_all_tags(self.admin_image_client)
        self.assertEqual([], metadef_namespace.tags)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.key_manager.v1 import secret_acl
from openstack.tests.unit import base


PUT_PATCH_EXAMPLE = {"acl_ref": "https://barbican/v1/secrets/UUID/acl"}
GET_EXAMPLE = {
    "read": {
        "updated": "2015-05-12T20:08:47.644264",
        "created": "2015-05-12T19:23:44.019168",
        "users": ["u1", "u2"],
        "project-access": True,
    }
}


class TestSecretACL(base.TestCase):
    def test_basic(self):
        sot = secret_acl.SecretACL(secret_id="SID")
        self.assertEqual('/secrets/%(secret_id)s/acl', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        # GET-shape payload
        sot = secret_acl.SecretACL(**GET_EXAMPLE)
        self.assertIn("project-access", sot.read)
        self.assertEqual(["u1", "u2"], sot.read["users"])

        # PUT/PATCH-shape payload
        sot = secret_acl.SecretACL(**PUT_PATCH_EXAMPLE)
        self.assertEqual(PUT_PATCH_EXAMPLE["acl_ref"], sot.acl_ref)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.dns.v2 import tld
from openstack.tests.unit import base

IDENTIFIER = "NAME"
EXAMPLE = {
    "id": IDENTIFIER,
    "name": "com",
    "description": "tld description",
}


class TestTLD(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.resp.status_code = 200
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.post = mock.Mock(return_value=self.resp)
        self.sess.default_microversion = None

    def test_basic(self):
        sot = tld.TLD()
        self.assertEqual(None, sot.resource_key)
        self.assertEqual("tlds", sot.resources_key)
        self.assertEqual("/tlds", sot.base_path)
        self.assertTrue(sot.allow_list)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)

        self.assertEqual("PATCH", sot.commit_method)

        self.assertDictEqual(
            {
                "description": "description",
                "name": "name",
                "limit": "limit",
                "marker": "marker",
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = tld.TLD(**EXAMPLE)
        self.assertEqual(IDENTIFIER, sot.id)
        self.assertEqual(EXAMPLE["description"], sot.description)
        self.assertEqual(EXAMPLE["name"], sot.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Generator
from typing import Any, Self
import uuid

from keystoneauth1 import adapter

from openstack import resource


class MessageResource(resource.Resource):
    # FIXME(anyone): The name string of `location` field of Zaqar API response
    # is lower case. That is inconsistent with the guide from API-WG. This is
    # a workaround for this issue.
    location = resource.Header("location")

    #: The ID to identify the client accessing Zaqar API. Must be specified
    #: in header for each API request.
    client_id = resource.Header("Client-ID")
    #: The ID to identify the project. Must be provided when keystone
    #: authentication is not enabled in Zaqar service.
    project_id = resource.Header("X-PROJECT-ID")

    @classmethod
    def list(
        cls,
        session: adapter.Adapter,
        paginated: bool = True,
        base_path: str | None = None,
        allow_unknown_params: bool = False,
        *,
        microversion: str | None = None,
        headers: dict[str, str] | None = None,
        max_items: int | None = None,
        **params: Any,
    ) -> Generator[Self, None, None]:
        """This method is a generator which yields resource objects.

        This is almost the copy of list method of resource.Resource class.
        The only difference is the request header now includes `Client-ID`
        and `X-PROJECT-ID` fields which are required by Zaqar v2 API.
        """
        more_data = True

        if base_path is None:
            base_path = cls.base_path

        uri = base_path % params

        project_id = params.get('project_id', None) or session.get_project_id()
        assert project_id is not None

        headers = {
            "Client-ID": params.get('client_id', None) or str(uuid.uuid4()),
            "X-PROJECT-ID": project_id,
        }

        query_params = cls._query_mapping._transpose(params, cls)
        while more_data:
            resp = session.get(
                uri, headers=headers, params=query_params
            ).json()[cls.resources_key]

            if not resp:
                more_data = False

            yielded = 0
            new_marker = None
            for data in resp:
                value = cls.existing(**data)
                new_marker = value.id
                yielded += 1
                yield value

            if not paginated:
                return
            if "limit" in query_params and yielded < query_params["limit"]:
                return
            query_params["limit"] = yielded
            query_params["marker"] = new_marker

    def fetch(
        self,
        session: adapter.Adapter,
        requires_id: bool = True,
        base_path: str | None = None,
        error_message: str | None = None,
        skip_cache: bool = False,
        **kwargs: Any,
    ) -> Self:
        request = self._prepare_request(
            requires_id=requires_id, base_path=base_path
        )
        headers: dict[str, str] = {
            "Client-ID": self.client_id or str(uuid.uuid4()),
            "X-PROJECT-ID": self.project_id or session.get_project_id() or "",
        }
        request.headers.update(headers)
        response = session.get(
            request.url, headers=headers, skip_cache=skip_cache
        )
        self._translate_response(response)

        return self

    @classmethod
    def _transform_create_request(
        cls,
        session: adapter.Adapter,
        request: resource._Request,
        *,
        microversion: str | None,
    ) -> resource._Request:
        if not request.headers.get('Client-ID'):
            request.headers['Client-ID'] = str(uuid.uuid4())
        if not request.headers.get('X-PROJECT-ID'):
            request.headers['X-PROJECT-ID'] = session.get_project_id() or ""
        return request

    def delete(
        self,
        session: adapter.Adapter,
        error_message: str | None = None,
        *,
        microversion: str | None = None,
        **kwargs: Any,
    ) -> Self:
        request = self._prepare_request()
        headers: dict[str, str] = {
            "Client-ID": self.client_id or str(uuid.uuid4()),
            "X-PROJECT-ID": self.project_id or session.get_project_id() or "",
        }
        request.headers.update(headers)
        response = session.delete(request.url, headers=headers)

        self._translate_response(response, has_body=False)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Self

from keystoneauth1 import adapter

from openstack.baremetal.v1 import _common
from openstack import exceptions
from openstack import resource
from openstack import utils


class Allocation(resource.Resource):
    resources_key = 'allocations'
    base_path = '/allocations'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_patch = True
    commit_method = 'PATCH'

    _query_mapping = resource.QueryParameters(
        'node',
        'resource_class',
        'state',
        'owner',
        fields={'type': _common.fields_type},
    )

    # Allocation update is available since 1.57
    # Backfilling allocations is available since 1.58
    # owner attribute is available since 1.60
    _max_microversion = '1.60'

    #: The candidate nodes for this allocation.
    candidate_nodes = resource.Body('candidate_nodes', type=list)
    #: Timestamp at which the allocation was created.
    created_at = resource.Body('created_at')
    #: A set of one or more arbitrary metadata key and value pairs.
    extra = resource.Body('extra', type=dict)
    #: The UUID for the allocation.
    id = resource.Body('uuid', alternate_id=True)
    #: The last error for the allocation.
    last_error = resource.Body("last_error")
    #: A list of relative links, including the self and bookmark links.
    links = resource.Body('links', type=list)
    #: The name of the allocation.
    name = resource.Body('name')
    #: The node UUID or name to create the allocation against,
    #: bypassing the normal allocation process.
    node = resource.Body('node')
    #: UUID of the node this allocation belongs to.
    node_id = resource.Body('node_uuid')
    #: The tenant who owns the object
    owner = resource.Body('owner')
    #: The requested resource class.
    resource_class = resource.Body('resource_class')
    #: The state of the allocation.
    state = resource.Body('state')
    #: The requested traits.
    traits = resource.Body('traits', type=list)
    #: Timestamp at which the allocation was last updated.
    updated_at = resource.Body('updated_at')

    def wait(
        self,
        session: adapter.Adapter,
        timeout: float | None = None,
        ignore_error: bool = False,
    ) -> Self:
        """Wait for the allocation to become active.

        :param session: The session to use for making this request.
        :param timeout: How much (in seconds) to wait for the allocation.
            The value of ``None`` (the default) means no client-side timeout.
        :param ignore_error: If ``True``, this call will raise an exception
            if the allocation reaches the ``error`` state. Otherwise the error
            state is considered successful and the call returns.

        :return: This :class:`Allocation` instance.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if allocation
            fails and ``ignore_error`` is ``False``.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` on timeout.
        """
        if self.state == 'active':
            return self

        for count in utils.iterate_timeout(
            timeout, f"Timeout waiting for the allocation {self.id}"
        ):
            self.fetch(session)

            if self.state == 'error' and not ignore_error:
                raise exceptions.ResourceFailure(
                    f"Allocation {self.id} failed: {self.last_error}"
                )
            elif self.state != 'allocating':
                return self

            session.log.debug(  # type: ignore[attr-defined]
                'Still waiting for the allocation %(allocation)s '
                'to become active, the current state is %(state)s',
                {'allocation': self.id, 'state': self.state},
            )

        assert False, "unreachable"



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# mypy: disable-error-code="method-assign"

from openstack.network.v2 import network
from openstack.network.v2 import port
from openstack.network.v2 import subnet
from openstack.tests.functional.network.v2 import common


class TestPort(common.TestTagNeutron):
    IPV4 = 4
    CIDR = "10.100.0.0/24"
    NET_ID: str
    SUB_ID: str
    PORT_ID: str

    def setUp(self):
        super().setUp()
        self.NET_NAME = self.getUniqueString()
        self.SUB_NAME = self.getUniqueString()
        self.PORT_NAME = self.getUniqueString()
        self.UPDATE_NAME = self.getUniqueString()
        net = self.user_cloud.network.create_network(name=self.NET_NAME)
        assert isinstance(net, network.Network)
        self.assertEqual(self.NET_NAME, net.name)
        self.NET_ID = net.id
        sub = self.user_cloud.network.create_subnet(
            name=self.SUB_NAME,
            ip_version=self.IPV4,
            network_id=self.NET_ID,
            cidr=self.CIDR,
        )
        assert isinstance(sub, subnet.Subnet)
        self.assertEqual(self.SUB_NAME, sub.name)
        self.SUB_ID = sub.id
        prt = self.user_cloud.network.create_port(
            name=self.PORT_NAME, network_id=self.NET_ID
        )
        assert isinstance(prt, port.Port)
        self.assertEqual(self.PORT_NAME, prt.name)
        self.PORT_ID = self.ID = prt.id
        self.get_command = self.user_cloud.network.get_port

    def tearDown(self):
        sot = self.user_cloud.network.delete_port(
            self.PORT_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_subnet(
            self.SUB_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_network(
            self.NET_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        super().tearDown()

    def test_find(self):
        sot = self.user_cloud.network.find_port(
            self.PORT_NAME, ignore_missing=False
        )
        self.assertEqual(self.PORT_ID, sot.id)

    def test_get(self):
        sot = self.user_cloud.network.get_port(self.PORT_ID)
        self.assertEqual(self.PORT_ID, sot.id)
        self.assertEqual(self.PORT_NAME, sot.name)
        self.assertEqual(self.NET_ID, sot.network_id)

    def test_list(self):
        ids = [o.id for o in self.user_cloud.network.ports()]
        self.assertIn(self.PORT_ID, ids)

    def test_update(self):
        assert self.PORT_ID is not None
        sot = self.user_cloud.network.update_port(
            self.PORT_ID, name=self.UPDATE_NAME
        )
        self.assertEqual(self.UPDATE_NAME, sot.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from typing import Any

from keystoneauth1 import adapter
import requests

from openstack import exceptions
from openstack import resource
from openstack import utils


class ShareReplica(resource.Resource):
    resource_key = "share_replica"
    resources_key = "share_replicas"
    base_path = "share-replicas"

    # capbilities
    allow_create = True
    allow_fetch = True
    allow_commit = False
    allow_delete = True
    allow_list = True
    allow_head = False

    #: properties
    #: The UUID of the share from which to create a share replica.
    share_id = resource.Body("share_id", type=str)
    #: The status of a share replica. List of possible values: available,
    #: error, creating, deleting, or error_deleting.
    status = resource.Body("status", type=str)
    #: If the share replica has its cast_rules_to_readonly attribute set
    #: to True, all existing access rules will be cast to read/only.
    cast_rules_to_readonly = resource.Body("cast_rules_to_readonly", type=bool)
    #: The date and time stamp when the resource was last updated within the
    #: service's database. If a resource was never updated after it was
    #: created, the value of this parameter is set to null.
    updated_at = resource.Body("updated_at", type=str)
    #: The share network ID where the resource is exported to.
    share_network_id = resource.Body("share_network_id", type=str)
    #: The UUID of the share server.
    share_server_id = resource.Body("share_server_id", type=str)
    #: The host name of the share replica.
    host = resource.Body("host", type=str)
    #: The replica state of a share replica.
    replica_state = resource.Body("replica_state", type=str)
    #: The date and time stamp when the resource was created within the
    #: service's database.
    created_at = resource.Body("created_at", type=str)

    def _action(
        self,
        session: adapter.Adapter,
        body: dict[str, Any],
        microversion: str | None = None,
    ) -> requests.Response:
        """Preform server actions given the message body."""
        url = utils.urljoin(self.base_path, self.id, 'action')
        headers = {'Accept': ''}
        microversion = microversion or self._get_microversion(session)
        response = session.post(
            url, json=body, headers=headers, microversion=microversion
        )
        exceptions.raise_from_response(response)
        return response

    def reset_status(
        self, session: adapter.Adapter, status: str
    ) -> requests.Response:
        """Reset the status of the share replica"""
        body = {'reset_status': {'status': status}}
        return self._action(session, body)

    def reset_replica_state(
        self, session: adapter.Adapter, replica_state: str
    ) -> requests.Response:
        """Reset replica_state of the share replica"""
        body = {'reset_replica_state': {'replica_state': replica_state}}
        return self._action(session, body)

    def force_delete(
        self, session: adapter.Adapter, *, ignore_missing: bool = False
    ) -> requests.Response:
        """Force-delete share replica"""
        try:
            body = {'force_delete': None}
        except exceptions.ResourceNotFound:
            if not ignore_missing:
                raise
        return self._action(session, body)

    def promote(self, session: adapter.Adapter) -> requests.Response:
        """Promote share replica"""
        body = {'promote': None}
        return self._action(session, body)

    def resync(self, session: adapter.Adapter) -> requests.Response:
        """Resync share replica"""
        body = {'resync': None}
        return self._action(session, body)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.network.v2 import (
    qos_minimum_bandwidth_rule as _qos_minimum_bandwidth_rule,
)
from openstack.tests.functional import base


class TestQoSMinimumBandwidthRule(base.BaseFunctionalTest):
    QOS_POLICY_ID: str
    QOS_IS_SHARED = False
    QOS_POLICY_DESCRIPTION = "QoS policy description"
    RULE_ID: str
    RULE_MIN_KBPS = 1200
    RULE_MIN_KBPS_NEW = 1800
    RULE_DIRECTION = "egress"

    def setUp(self):
        super().setUp()

        if not self.operator_cloud:
            self.skipTest("Operator cloud is required for this test")

        # Skip the tests if qos-bw-limit-direction extension is not enabled.
        if not self.operator_cloud.network.find_extension(
            "qos-bw-limit-direction"
        ):
            self.skipTest("Network qos-bw-limit-direction extension disabled")

        self.QOS_POLICY_NAME = self.getUniqueString()
        qos_policy = self.operator_cloud.network.create_qos_policy(
            description=self.QOS_POLICY_DESCRIPTION,
            name=self.QOS_POLICY_NAME,
            shared=self.QOS_IS_SHARED,
        )
        self.QOS_POLICY_ID = qos_policy.id
        qos_min_bw_rule = (
            self.operator_cloud.network.create_qos_minimum_bandwidth_rule(
                self.QOS_POLICY_ID,
                direction=self.RULE_DIRECTION,
                min_kbps=self.RULE_MIN_KBPS,
            )
        )
        assert isinstance(
            qos_min_bw_rule,
            _qos_minimum_bandwidth_rule.QoSMinimumBandwidthRule,
        )
        self.assertEqual(self.RULE_MIN_KBPS, qos_min_bw_rule.min_kbps)
        self.assertEqual(self.RULE_DIRECTION, qos_min_bw_rule.direction)
        self.RULE_ID = qos_min_bw_rule.id

    def tearDown(self):
        rule = self.operator_cloud.network.delete_qos_minimum_bandwidth_rule(
            self.RULE_ID, self.QOS_POLICY_ID
        )
        qos_policy = self.operator_cloud.network.delete_qos_policy(
            self.QOS_POLICY_ID
        )
        self.assertIsNone(rule)
        self.assertIsNone(qos_policy)
        super().tearDown()

    def test_find(self):
        sot = self.operator_cloud.network.find_qos_minimum_bandwidth_rule(
            self.RULE_ID, self.QOS_POLICY_ID, ignore_missing=False
        )
        self.assertEqual(self.RULE_ID, sot.id)
        self.assertEqual(self.RULE_DIRECTION, sot.direction)
        self.assertEqual(self.RULE_MIN_KBPS, sot.min_kbps)

    def test_get(self):
        sot = self.operator_cloud.network.get_qos_minimum_bandwidth_rule(
            self.RULE_ID, self.QOS_POLICY_ID
        )
        self.assertEqual(self.RULE_ID, sot.id)
        self.assertEqual(self.QOS_POLICY_ID, sot.qos_policy_id)
        self.assertEqual(self.RULE_DIRECTION, sot.direction)
        self.assertEqual(self.RULE_MIN_KBPS, sot.min_kbps)

    def test_list(self):
        rule_ids = [
            o.id
            for o in self.operator_cloud.network.qos_minimum_bandwidth_rules(
                self.QOS_POLICY_ID
            )
        ]
        self.assertIn(self.RULE_ID, rule_ids)

    def test_update(self):
        sot = self.operator_cloud.network.update_qos_minimum_bandwidth_rule(
            self.RULE_ID, self.QOS_POLICY_ID, min_kbps=self.RULE_MIN_KBPS_NEW
        )
        self.assertEqual(self.RULE_MIN_KBPS_NEW, sot.min_kbps)



# Copyright(c) 2022 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack import resource


class VMove(resource.Resource):
    resource_key = "vmove"
    resources_key = "vmoves"
    base_path = "/notifications/%(notification_id)s/vmoves"

    # capabilities
    # 1] GET /v1/notifications/{notification_uuid}/vmoves
    # 2] GET /v1/notifications/{notification_uuid}/vmoves/{vmove_uuid}
    allow_list = True
    allow_fetch = True

    _query_mapping = resource.QueryParameters(
        "sort_key",
        "sort_dir",
        "type",
        "status",
    )

    #: A ID of representing this vmove
    id = resource.Body("id")
    #: A UUID of representing this vmove
    uuid = resource.Body("uuid")
    #: The notification UUID this vmove belongs to(in URI)
    notification_id = resource.URI("notification_id")
    #: A created time of this vmove
    created_at = resource.Body("created_at")
    #: A latest updated time of this vmove
    updated_at = resource.Body("updated_at")
    #: The instance uuid of this vmove
    server_id = resource.Body("instance_uuid")
    #: The instance name of this vmove
    server_name = resource.Body("instance_name")
    #: The source host of this vmove
    source_host = resource.Body("source_host")
    #: The dest host of this vmove
    dest_host = resource.Body("dest_host")
    #: A start time of this vmove
    start_time = resource.Body("start_time")
    #: A end time of this vmove
    end_time = resource.Body("end_time")
    #: The status of this vmove
    status = resource.Body("status")
    #: The type of this vmove
    type = resource.Body("type")
    #: The message of this vmove
    message = resource.Body("message")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional import base
from openstack import utils


class TestService(base.BaseFunctionalTest):
    # listing services is slooowwww
    TIMEOUT_SCALING_FACTOR = 2

    def test_list(self):
        sot = list(self.operator_cloud.block_storage.services())
        self.assertIsNotNone(sot)

    def test_disable_enable(self):
        block_storage = utils.ensure_service_version(
            self.operator_cloud.block_storage, '3'
        )
        for srv in block_storage.services():
            # only nova-block_storage can be updated
            if srv.name == 'nova-block_storage':
                block_storage.disable_service(srv)
                block_storage.enable_service(srv)
                break

    def test_find(self):
        for srv in self.operator_cloud.block_storage.services():
            self.operator_cloud.block_storage.find_service(
                srv.name,
                host=srv.host,
                ignore_missing=False,
            )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Policy(resource.Resource):
    resource_key = 'policy'
    resources_key = 'policies'
    base_path = '/policies'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    commit_method = 'PATCH'

    # Properties
    #: The policy rule set itself, as a serialized blob. *Type: string*
    blob = resource.Body('blob')
    #: The links for the policy resource.
    links = resource.Body('links')
    #: The ID for the project.
    project_id = resource.Body('project_id')
    #: The MIME Media Type of the serialized policy blob. *Type: string*
    type = resource.Body('type')
    #: The ID of the user who owns the policy
    user_id = resource.Body('user_id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Delete resources from the DNS service.

For a full guide see
https://docs.openstack.org/openstacksdk/latest/user/guides/dns.html
"""


def delete_zone(conn, name_or_id):
    print(f"Delete Zone: {name_or_id}")

    zone = conn.dns.find_zone(name_or_id)

    if zone:
        conn.dns.delete_zone(zone.id)
    else:
        return None


def delete_recordset(conn, name_or_id, recordset_name):
    print(f"Deleting Recordset: {recordset_name} in Zone: {name_or_id}")

    zone = conn.dns.find_zone(name_or_id)

    if zone:
        try:
            recordset = conn.dns.find_recordset(zone.id, recordset_name)
            if recordset:
                conn.dns.delete_recordset(recordset, zone.id)
            else:
                print("Recordset not found")
        except Exception as e:
            print(f"{e}")
    else:
        return None



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class MetadefObject(resource.Resource):
    resources_key = 'objects'
    base_path = '/metadefs/namespaces/%(namespace_name)s/objects'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        "visibility",
        "resource_types",
        "sort_key",
        "sort_dir",
    )

    created_at = resource.Body('created_at')
    description = resource.Body('description')
    name = resource.Body('name', alternate_id=True)
    namespace_name = resource.URI('namespace_name')
    properties = resource.Body('properties')
    required = resource.Body('required')
    updated_at = resource.Body('updated_at')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testtools import matchers

from openstack import exceptions
from openstack.tests.unit import base


class TestRoleAssignment(base.TestCase):
    IS_INHERITED = False

    def _build_role_assignment_response(
        self, role_id, scope_type, scope_id, entity_type, entity_id
    ):
        self.assertThat(['group', 'user'], matchers.Contains(entity_type))
        self.assertThat(['project', 'domain'], matchers.Contains(scope_type))
        # NOTE(notmorgan): Links are thrown out by shade, but we construct them
        # for corectness.
        link_str = (
            'https://identity.example.com/identity/v3/{scope_t}s'
            '/{scopeid}/{entity_t}s/{entityid}/roles/{roleid}'
        )
        return [
            {
                'links': {
                    'assignment': link_str.format(
                        scope_t=scope_type,
                        scopeid=scope_id,
                        entity_t=entity_type,
                        entityid=entity_id,
                        roleid=role_id,
                    )
                },
                'role': {'id': role_id},
                'scope': {scope_type: {'id': scope_id}},
                entity_type: {'id': entity_id},
            }
        ]

    def setUp(self):
        super().setUp()
        self.role_data = self._get_role_data()
        self.domain_data = self._get_domain_data()
        self.user_data = self._get_user_data(
            domain_id=self.domain_data.domain_id
        )
        self.project_data = self._get_project_data(
            domain_id=self.domain_data.domain_id
        )
        self.project_data_v2 = self._get_project_data(
            project_name=self.project_data.project_name,
            project_id=self.project_data.project_id,
            v3=False,
        )
        self.group_data = self._get_group_data(
            domain_id=self.domain_data.domain_id
        )

        self.user_project_assignment = self._build_role_assignment_response(
            role_id=self.role_data.role_id,
            scope_type='project',
            scope_id=self.project_data.project_id,
            entity_type='user',
            entity_id=self.user_data.user_id,
        )

        self.group_project_assignment = self._build_role_assignment_response(
            role_id=self.role_data.role_id,
            scope_type='project',
            scope_id=self.project_data.project_id,
            entity_type='group',
            entity_id=self.group_data.group_id,
        )

        self.user_domain_assignment = self._build_role_assignment_response(
            role_id=self.role_data.role_id,
            scope_type='domain',
            scope_id=self.domain_data.domain_id,
            entity_type='user',
            entity_id=self.user_data.user_id,
        )

        self.group_domain_assignment = self._build_role_assignment_response(
            role_id=self.role_data.role_id,
            scope_type='domain',
            scope_id=self.domain_data.domain_id,
            entity_type='group',
            entity_id=self.group_data.group_id,
        )

        # Cleanup of instances to ensure garbage collection/no leaking memory
        # in tests.
        self.addCleanup(delattr, self, 'role_data')
        self.addCleanup(delattr, self, 'user_data')
        self.addCleanup(delattr, self, 'domain_data')
        self.addCleanup(delattr, self, 'group_data')
        self.addCleanup(delattr, self, 'project_data')
        self.addCleanup(delattr, self, 'project_data_v2')
        self.addCleanup(delattr, self, 'user_project_assignment')
        self.addCleanup(delattr, self, 'group_project_assignment')
        self.addCleanup(delattr, self, 'user_domain_assignment')
        self.addCleanup(delattr, self, 'group_domain_assignment')

    def get_mock_url(
        self,
        service_type='identity',
        interface='public',
        resource='role_assignments',
        append=None,
        base_url_append='v3',
        qs_elements=None,
        inherited=False,
    ):
        if inherited:
            base_url_append = base_url_append + '/OS-INHERIT'
        if append and inherited:
            append.append('inherited_to_projects')

        return super().get_mock_url(
            service_type,
            interface,
            resource,
            append,
            base_url_append,
            qs_elements,
        )

    def __get(
        self, resource, data, attr, qs_elements, use_name=False, is_found=True
    ):
        if not use_name:
            if is_found:
                return [
                    dict(
                        method='GET',
                        uri=self.get_mock_url(
                            resource=resource + 's',  # do roles from role
                            append=[getattr(data, attr)],
                            qs_elements=qs_elements,
                        ),
                        status_code=200,
                        json=data.json_response,
                    )
                ]
            else:
                return [
                    dict(
                        method='GET',
                        uri=self.get_mock_url(
                            resource=resource + 's',  # do roles from role
                            append=[getattr(data, attr)],
                            qs_elements=qs_elements,
                        ),
                        status_code=404,
                    ),
                    dict(
                        method='GET',
                        uri=self.get_mock_url(
                            resource=resource + 's',  # do roles from role
                            qs_elements=qs_elements,
                        ),
                        status_code=200,
                        json={(resource + 's'): []},
                    ),
                ]
        else:
            return [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource=resource + 's',
                        append=[getattr(data, attr)],
                        qs_elements=qs_elements,
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource=resource + 's',
                        qs_elements=[
                            'name=' + getattr(data, attr),
                            *qs_elements,
                        ],
                    ),
                    status_code=200,
                    json={(resource + 's'): [data.json_response[resource]]},
                ),
            ]

    def __user_mocks(
        self, user_data, use_name, is_found=True, domain_data=None
    ):
        qs_elements = []
        if domain_data:
            qs_elements = ['domain_id=' + domain_data.domain_id]
        uri_mocks = []
        if not use_name:
            uri_mocks.append(
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='users',
                        append=[user_data.user_id],
                        # TODO(stephenfin): We shouldn't be passing domain ID
                        # here since it's unnecessary, but that requires a much
                        # larger rework of the Resource.find method.
                        qs_elements=qs_elements,
                    ),
                    json=user_data.json_response if is_found else None,
                    status_code=200 if is_found else 404,
                ),
            )
        else:
            uri_mocks += [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='users',
                        append=[user_data.name],
                        qs_elements=qs_elements,
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='users',
                        qs_elements=[*qs_elements, 'name=' + user_data.name],
                    ),
                    status_code=200,
                    json={
                        'users': (
                            [user_data.json_response['user']]
                            if is_found
                            else []
                        )
                    },
                ),
            ]
        return uri_mocks

    def _get_mock_role_query_urls(
        self,
        role_data,
        domain_data=None,
        project_data=None,
        group_data=None,
        user_data=None,
        use_role_name=False,
        use_domain_name=False,
        use_project_name=False,
        use_group_name=False,
        use_user_name=False,
        use_domain_in_query=True,
    ):
        """Build uri mocks for querying role assignments"""
        uri_mocks = []

        if domain_data:
            uri_mocks.extend(
                self.__get(
                    'domain',
                    domain_data,
                    'domain_id' if not use_domain_name else 'domain_name',
                    [],
                    use_name=use_domain_name,
                )
            )

        qs_elements = []
        if domain_data and use_domain_in_query:
            qs_elements = ['domain_id=' + domain_data.domain_id]

        uri_mocks.extend(
            self.__get(
                'role',
                role_data,
                'role_id' if not use_role_name else 'role_name',
                [],
                use_name=use_role_name,
            )
        )

        if user_data:
            uri_mocks.extend(
                self.__user_mocks(
                    user_data,
                    use_user_name,
                    is_found=True,
                    domain_data=domain_data,
                )
            )

        if group_data:
            uri_mocks.extend(
                self.__get(
                    'group',
                    group_data,
                    'group_id' if not use_group_name else 'group_name',
                    qs_elements,
                    use_name=use_group_name,
                )
            )

        if project_data:
            uri_mocks.extend(
                self.__get(
                    'project',
                    project_data,
                    'project_id' if not use_project_name else 'project_name',
                    qs_elements,
                    use_name=use_project_name,
                )
            )

        return uri_mocks

    def test_grant_role_user_id_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
            use_role_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.user_id,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_user_name_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
            use_role_name=True,
            use_user_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )

    def test_grant_role_user_id_project_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.grant_role(
                self.role_data.role_id,
                user=self.user_data.user_id,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_user_name_project_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
            use_role_name=True,
            use_user_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_group_id_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            group_data=self.group_data,
            use_role_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_id,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_group_name_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            group_data=self.group_data,
            use_role_name=True,
            use_group_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_group_id_project_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            group_data=self.group_data,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.grant_role(
                self.role_data.role_id,
                group=self.group_data.group_id,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_group_name_project_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            group_data=self.group_data,
            use_role_name=True,
            use_group_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    # ===== Domain
    def test_grant_role_user_id_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            user_data=self.user_data,
            use_role_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.user_id,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_user_name_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            user_data=self.user_data,
            use_role_name=True,
            use_user_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )

    def test_grant_role_user_id_domain_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            user_data=self.user_data,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.grant_role(
                self.role_data.role_id,
                user=self.user_data.user_id,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_user_name_domain_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            user_data=self.user_data,
            use_role_name=True,
            use_user_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_group_id_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            group_data=self.group_data,
            use_role_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_id,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_group_name_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            group_data=self.group_data,
            use_role_name=True,
            use_group_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_group_id_domain_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            group_data=self.group_data,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.grant_role(
                self.role_data.role_id,
                group=self.group_data.group_id,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_role_group_name_domain_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            group_data=self.group_data,
            use_role_name=True,
            use_group_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    # ==== Revoke
    def test_revoke_role_user_id_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
            use_role_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.user_id,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_user_name_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
            use_role_name=True,
            use_user_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )

    def test_revoke_role_user_id_project_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.revoke_role(
                self.role_data.role_id,
                user=self.user_data.user_id,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_user_name_project_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
            use_role_name=True,
            use_user_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_group_id_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            group_data=self.group_data,
            use_role_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_id,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_group_name_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            group_data=self.group_data,
            use_role_name=True,
            use_group_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_group_id_project_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            group_data=self.group_data,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.revoke_role(
                self.role_data.role_id,
                group=self.group_data.group_id,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_group_name_project_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            group_data=self.group_data,
            use_role_name=True,
            use_group_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                project=self.project_data.project_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    # ==== Domain
    def test_revoke_role_user_id_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            user_data=self.user_data,
            use_role_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.user_id,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_user_name_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            user_data=self.user_data,
            use_role_name=True,
            use_user_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )

    def test_revoke_role_user_id_domain_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            user_data=self.user_data,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.revoke_role(
                self.role_data.role_id,
                user=self.user_data.user_id,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_user_name_domain_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            user_data=self.user_data,
            use_role_name=True,
            use_user_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_group_id_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            group_data=self.group_data,
            use_role_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_id,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_group_name_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            group_data=self.group_data,
            use_role_name=True,
            use_group_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_group_id_domain_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            group_data=self.group_data,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.revoke_role(
                self.role_data.role_id,
                group=self.group_data.group_id,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_revoke_role_group_name_domain_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data,
            group_data=self.group_data,
            use_role_name=True,
            use_group_name=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='domains',
                        append=[
                            self.domain_data.domain_id,
                            'groups',
                            self.group_data.group_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertFalse(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_id,
                inherited=self.IS_INHERITED,
            )
        )
        self.assert_calls()

    def test_grant_no_role(self):
        uris = self.__get(
            'domain', self.domain_data, 'domain_name', [], use_name=True
        )
        uris.extend(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='roles',
                        append=[self.role_data.role_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='roles',
                        qs_elements=[
                            'name=' + self.role_data.role_name,
                        ],
                    ),
                    status_code=200,
                    json={'roles': []},
                ),
            ]
        )
        self.register_uris(uris)

        with self.assertRaises(exceptions.NotFoundException):
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_name,
                inherited=self.IS_INHERITED,
            )
        self.assert_calls()

    def test_revoke_no_role(self):
        uris = self.__get(
            'domain', self.domain_data, 'domain_name', [], use_name=True
        )
        uris.extend(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='roles',
                        append=[self.role_data.role_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='roles',
                        qs_elements=[
                            'name=' + self.role_data.role_name,
                        ],
                    ),
                    status_code=200,
                    json={'roles': []},
                ),
            ]
        )
        self.register_uris(uris)

        with self.assertRaises(exceptions.NotFoundException):
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_name,
                inherited=self.IS_INHERITED,
            )
        self.assert_calls()

    def test_grant_no_user_or_group_specified(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True
        )
        self.register_uris(uris)
        with self.assertRaises(
            exceptions.SDKException,
            msg='Must specify either a user or a group',
        ):
            self.cloud.grant_role(
                self.role_data.role_name,
                inherited=self.IS_INHERITED,
            )
        self.assert_calls()

    def test_revoke_no_user_or_group_specified(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True
        )
        self.register_uris(uris)
        with self.assertRaises(
            exceptions.SDKException,
            msg='Must specify either a user or a group',
        ):
            self.cloud.revoke_role(
                self.role_data.role_name,
                inherited=self.IS_INHERITED,
            )
        self.assert_calls()

    def test_grant_both_user_and_group(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True
        )
        self.register_uris(uris)

        with self.assertRaises(
            exceptions.SDKException,
            msg='Specify either a group or a user, not both',
        ):
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                group=self.group_data.group_name,
                inherited=self.IS_INHERITED,
            )
        self.assert_calls()

    def test_revoke_both_user_and_group(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True
        )
        uris.extend(self.__user_mocks(self.user_data, use_name=True))
        uris.extend(
            self.__get(
                'group', self.group_data, 'group_name', [], use_name=True
            )
        )
        self.register_uris(uris)

        with self.assertRaises(
            exceptions.SDKException,
            msg='Specify either a group or a user, not both',
        ):
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                group=self.group_data.group_name,
                inherited=self.IS_INHERITED,
            )

    def test_grant_both_project_and_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
            domain_data=self.domain_data,
            use_role_name=True,
            use_user_name=True,
            use_project_name=True,
            use_domain_name=True,
            use_domain_in_query=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_name,
                domain=self.domain_data.domain_name,
                inherited=self.IS_INHERITED,
            )
        )

    def test_revoke_both_project_and_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data,
            user_data=self.user_data,
            domain_data=self.domain_data,
            use_role_name=True,
            use_user_name=True,
            use_project_name=True,
            use_domain_name=True,
            use_domain_in_query=True,
        )
        uris.extend(
            [
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    complete_qs=True,
                    status_code=204,
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        resource='projects',
                        append=[
                            self.project_data.project_id,
                            'users',
                            self.user_data.user_id,
                            'roles',
                            self.role_data.role_id,
                        ],
                        inherited=self.IS_INHERITED,
                    ),
                    status_code=200,
                ),
            ]
        )
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_name,
                domain=self.domain_data.domain_name,
                inherited=self.IS_INHERITED,
            )
        )

    def test_grant_no_project_or_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            use_role_name=True,
        )

        self.register_uris(uris)

        with self.assertRaises(
            exceptions.SDKException,
            msg='Must specify either a domain, project or system',
        ):
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                inherited=self.IS_INHERITED,
            )
        self.assert_calls()

    def test_revoke_no_project_or_domain_or_system(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            use_role_name=True,
        )

        self.register_uris(uris)

        with self.assertRaises(
            exceptions.SDKException,
            msg='Must specify either a domain, project or system',
        ):
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                inherited=self.IS_INHERITED,
            )
        self.assert_calls()

    def test_grant_bad_domain_exception(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='domains', append=['baddomain']
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='domains', qs_elements=['name=baddomain']
                    ),
                    status_code=404,
                ),
            ]
        )
        with self.assertRaises(exceptions.NotFoundException):
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain='baddomain',
                inherited=self.IS_INHERITED,
            )
        self.assert_calls()

    def test_revoke_bad_domain_exception(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='domains', append=['baddomain']
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource='domains', qs_elements=['name=baddomain']
                    ),
                    status_code=404,
                ),
            ]
        )
        with self.assertRaises(exceptions.NotFoundException):
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain='baddomain',
                inherited=self.IS_INHERITED,
            )
        self.assert_calls()


class TestInheritedRoleAssignment(TestRoleAssignment):
    IS_INHERITED = True
